from blinker import Signal
import requests
from requests.adapters import HTTPAdapter
from requests.auth import HTTPBasicAuth
from requests.packages.urllib3.poolmanager import PoolManager
import ssl

from . import __version__, __title__
from .services import services
from .exceptions import (
    InvalidTokenError,
    AuthenticationError,
    MediaAmpError,
    wrap_http_error,
    raise_for_json_exception,
)


SIGN_IN_URL = 'https://identity.auth.theplatform.{tld}/idm/web/Authentication/signIn'
REGISTRY_URL = 'https://access.auth.theplatform.{tld}/web/Registry/resolveDomain'


class Session(object):

    def __init__(self,
                 username,
                 password,
                 account_id,
                 auth_token=None,
                 user_directory='mpx',
                 region='US1',
                 service_registry=None,
                 token_duration=43200000,       # 12 hours
                 token_idle_timeout=14400000,   # 4 hours
                 use_ssl=True,
                 ):

        self.username = username
        self.password = password
        self.account = account_id
        self.auth_token = auth_token
        self.user_directory = user_directory
        self.region = region
        self.token_duration = token_duration
        self.token_idle_timeout = token_idle_timeout
        self.use_ssl = use_ssl
        self.registry_url = REGISTRY_URL.format(tld=self.regional_tld)
        self.signin_url = SIGN_IN_URL.format(tld=self.regional_tld)
        self.post_sign_in = Signal()
        self._registry = service_registry
        self.session = requests.Session()
        self.session.mount('https://', TLS1Adapter())
        self.session.headers.update({
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'User-Agent': 'Python Mediaamp %s' % __version__,
        })

    @property
    def registry(self):
        if self._registry is None:
            self._registry = self.resolve_domain()
        return self._registry

    @property
    def regional_tld(self):
        return 'eu' if 'eu' in self.region.lower() else 'com'

    @property
    def signin_username(self):
        return self.user_directory + '/' + self.username

    def resolve_domain(self):
        resp = self.get(self.registry_url, params={
            'schema': '1.1',
            '_accountId': self.account,
        })
        try:
            return resp['resolveDomainResponse']
        except KeyError:
            raise MediaAmpError('Unexpected response loading registry.')

    def sign_in(self):
        self.auth_token = None
        self.session.auth = HTTPBasicAuth(self.signin_username, self.password)
        result = self.get(self.signin_url, is_signin_request=True, params={
            'schema': '1.0',
            '_duration': self.token_duration,
            '_idleTimeout': self.token_idle_timeout,
        })
        try:
            self.auth_token = result['signInResponse']['token']
        except KeyError:
            raise AuthenticationError('Could not retrieve token.')
        self.post_sign_in.send(self)

    def request_json(self, method, url, retry_sign_in=True, is_signin_request=False, **kwargs):
        """ Requests JSON content from the supplied URL.

        This is the primary function to be used to make requests to the MPX API.
        Not only does it ensure that the body of the response can be encoded as
        Python via JSON, it also wraps exceptions and will auto-login using
        the supplied credentials when necessary (e.g. when a token expires).

        This API is known to return 200 statuses for requests that fail. It's
        their convention to include the HTTP response code in the body of
        the JSON returned. This checks for that case and turns them into actual
        exceptions.

        """
        if self.auth_token is not None:
            self.session.auth = HTTPBasicAuth('', self.auth_token)
        elif not is_signin_request:
            self.sign_in()

        response = getattr(self.session, method)(url, **kwargs)

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            wrap_http_error(e)

        try:
            data = response.json()
        except ValueError:
            raise MediaAmpError('Response body can not be read as JSON. ')

        try:
            raise_for_json_exception(data)
        except InvalidTokenError:
            if retry_sign_in:
                self.sign_in()
                return self.request_json(method, url, retry_sign_in=False, **kwargs)
            else:
                raise

        return data

    def get(self, url, **kwargs):
        return self.request_json('get', url, **kwargs)

    def put(self, url, **kwargs):
        return self.request_json('put', url, **kwargs)

    def post(self, url, **kwargs):
        return self.request_json('post', url, **kwargs)

    def delete(self, url, **kwargs):
        return self.request_json('delete', url, **kwargs)

    def __getitem__(self, key):
        url = self.registry.get(key)
        if url is None:
            url = self.registry(key + ' read-only')
        if url is None:
            raise KeyError(key + ' not available.')
        if self.use_ssl:
            url = url.replace('http://', 'https://')
        return services[key](self, url)


class TLS1Adapter(HTTPAdapter):
    """ Force requests SSL to use TLS 1.

    As of 4/30/2015 thePlatform APIs use TLS 1.2 which causes EOF
    errors with the default adapter. They are planning to update to
    SHA-256 certs @ June 15. see:

    https://help.theplatform.com/display/trc/2015+certificate+security+upgrade

    """

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = PoolManager(
            num_pools=connections,
            maxsize=maxsize,
            block=block,
            ssl_version=ssl.PROTOCOL_TLSv1
        )



from datetime import datetime
from time import mktime


def decode_datetime(dt_in_millis):
    return datetime.utcfromtimestamp(dt_in_millis / 1000)


def encode_datetime(dt):
    return int(mktime(dt.utctimetuple()) * 1000)



from collections import defaultdict


class MediaAmpError(Exception):
    """Unspecified error occurred."""


class ClientError(MediaAmpError):
    """Bad request or unauthorized request."""


class ServerError(MediaAmpError):
    """Remote server error."""


class AuthenticationError(MediaAmpError):
    """Unable to authenticate with supplied credentials."""


class InvalidTokenError(AuthenticationError):
    """Supplied authentication token is not valid or has expired."""


class ServiceNotAvailable(MediaAmpError):
    """Could not find service by name in the registry."""


def wrap_http_error(error):
    response = error.response
    error_class = http_status_map[response.status_code]
    raise error_class(response.text)


http_status_map = defaultdict(lambda: MediaAmpError)
for code in range(400, 500):
    http_status_map[code] = ClientError
for code in (401, 403):
    http_status_map[code] = AuthenticationError
for code in range(500, 600):
    http_status_map[code] = ServerError


def raise_for_json_exception(data):
    if isinstance(data, dict) and data.get('isException'):
        status_code = data.get('responseCode', 400)
        description = data.get('description', 'Bad Request')
        title = data.get('title')
        if title in json_error_title_map:
            exc = json_error_title_map[title]
        elif 'invalid security token' in description.lower():
            exc = InvalidTokenError
        else:
            exc = http_status_map[status_code]
        raise exc(description)


json_error_title_map = {
    'com.theplatform.authentication.api.exception.InvalidTokenException': InvalidTokenError
}



from functools import partial
import re
from .exceptions import ServiceNotAvailable

services = {}

def register(cls):
    if issubclass(cls, BaseService):
        key = getattr(cls, 'registry_key', None)
        if not key:
            key = ' '.join(re.findall('[A-Z][^A-Z]*', cls.__name__))
        services[key] = cls


class Endpoint(object):

    def __init__(self, path=None, name=None, service=None, **kwargs):
        self.path = path
        self.name = name
        self.service = service
        self.default_params = kwargs.copy()
        self.default_params.setdefault('schema', '1.0')

    def urljoin(self, *args):
        parts = (self.service.base_url, self.path, self.name) + args
        return '/'.join([
            part.lstrip('/') for part in parts if part is not None
        ]).rstrip('/')

    def get(self, extra_path=None, **kwargs):
        return self._make_request('get', extra_path, **kwargs)

    def put(self, extra_path=None, **kwargs):
        return self._make_request('put', extra_path, **kwargs)

    def post(self, extra_path=None, **kwargs):
        return self._make_request('post', extra_path, **kwargs)

    def delete(self, extra_path=None, **kwargs):
        return self._make_request('delete', extra_path, **kwargs)

    def _make_request(self, method, extra_path=None, **kwargs):
        # merge default parameters with those supplied
        params = dict(self.default_params, **kwargs.pop('params', {}))
        extra_path = extra_path
        url = self.urljoin(extra_path)
        return self.service.session.request_json(method, url, params=params, **kwargs)

    def __call__(self, **kwargs):
        """ Override default URL parameters.

        Allow custom overrides of defaults to look like object
        initialization.

        """
        self.default_params.update(kwargs)
        return self


class BaseService(object):

    def __init__(self, session, base_url):
        self.session = session
        self.base_url = base_url
        self.init_endpoints()

    def init_endpoints(self):
        for k, v in self.__class__.__dict__.items():
            if isinstance(v, Endpoint):
                v.service = self
                v(account=self.session.account)
                if v.name is None:
                    v.name = k

    @property
    def Notifications(self):
        return Endpoint(name='notify', service=self, account=self.session.account)


DataEndpoint = partial(Endpoint, path='data')
BusinessEndpoint = partial(Endpoint, path='web')


@register
class AccessDataService(BaseService):
    Permission = DataEndpoint()
    Role = DataEndpoint()
    Authorization = BusinessEndpoint()
    Lookup = BusinessEndpoint()
    Registry = BusinessEndpoint()


@register
class AccountDataService(BaseService):
    Account = DataEndpoint()


@register
class CommerceConfigurationDataService(BaseService):
    CommerceRegistry = DataEndpoint()
    CheckoutConfiguration = DataEndpoint()
    FulfillmentConfiguration = DataEndpoint()
    PaymentConfiguration = DataEndpoint()
    Rule = DataEndpoint()
    RuleSet = DataEndpoint()
    TaxConfiguration = DataEndpoint()

@register
class CommerceEventDataService(BaseService):
    OrderTracking = DataEndpoint()


@register
class CuePointDataService(BaseService):
    CuePoint = Endpoint()
    CuePointType = Endpoint()


@register
class DeliveryDataService(BaseService):
    AccountSettings = DataEndpoint()
    AdPolicy = DataEndpoint()
    Restriction = DataEndpoint()
    UserAgent = DataEndpoint()


@register
class EndUserDataService(BaseService):
    Directory = DataEndpoint()
    Security = DataEndpoint()
    User = DataEndpoint()
    Authentication = BusinessEndpoint()
    Lookup = BusinessEndpoint()
    Self = BusinessEndpoint()


@register
class EntertainmentDataService(BaseService):
    AccountSettings = DataEndpoint()
    Channel = DataEndpoint()
    ChannelSchedule = DataEndpoint()
    Company = DataEndpoint()
    Credit = DataEndpoint()
    Listing = DataEndpoint()
    Location = DataEndpoint()
    Person = DataEndpoint()
    Program = DataEndpoint()
    ProgramAvailability = DataEndpoint()
    Station = DataEndpoint()
    Tag = DataEndpoint()
    TvSeason = DataEndpoint()


@register
class EntertainmentIngestDataService(BaseService):
    IngestConfig = DataEndpoint()
    IngestResult = DataEndpoint()
    Process = BusinessEndpoint()


@register
class EntertainmentFeedsService(BaseService):
    Feed = Endpoint(name='')


@register
class EntitlementDataService(BaseService):
    AccountSettings = DataEndpoint()
    Adapter = DataEndpoint()
    AdapterConfiguration = DataEndpoint()
    Device = DataEndpoint()
    DistributionRight = DataEndpoint()
    DistributionRightLicenseCount = DataEndpoint()
    Entitlement = DataEndpoint()
    LicenseCount = DataEndpoint()
    PhysicalDevice = DataEndpoint()
    ProductDevice = DataEndpoint()
    Rights = DataEndpoint()
    SubjectPolicy = DataEndpoint()
    UserDevice = DataEndpoint()


@register
class EntitlementWebService(BaseService):
    Entitlements = BusinessEndpoint()
    RegisterDevice = BusinessEndpoint()


@register
class EntitlementLicenseService(BaseService):
    ContentAccessRules = BusinessEndpoint(schema='1.2')
    License = BusinessEndpoint(schema='2.5')


@register
class FeedReaderDataService(BaseService):
    registry_key = 'FeedReader Data Service'
    FeedRecord = DataEndpoint()


@register
class FeedsDataService(BaseService):
    FeedAdapter = DataEndpoint()
    FeedConfig = DataEndpoint()


@register
class FeedsService(BaseService):
    Feed = Endpoint(name='')


@register
class FileManagementService(BaseService):
    FileManagement = BusinessEndpoint()


@register
class IngestDataService(BaseService):
    Adapter = DataEndpoint()
    AdapterConfiguration = DataEndpoint()
    Checksum = DataEndpoint()


@register
class IngestService(BaseService):
    ingest = Endpoint()
    test = Endpoint()


@register
class KeyDataService(BaseService):
    Key = DataEndpoint()
    UserKey = DataEndpoint()


@register
class LedgerDataService(BaseService):
    LedgerEntr = DataEndpoint()


@register
class LiveEventDataService(BaseService):
    LiveEncoder = DataEndpoint()
    LiveEvent = DataEndpoint()


@register
class LiveEventService(BaseService):
    Scheduling = BusinessEndpoint()


@register
class MediaDataService(BaseService):
    AccountSettings = DataEndpoint()
    AssetType = DataEndpoint()
    Category = DataEndpoint()
    Media = DataEndpoint()
    MediaDefaults = DataEndpoint()
    MediaFile = DataEndpoint()
    Provider = DataEndpoint()
    Release = DataEndpoint()
    Server = DataEndpoint()


@register
class MessageDataService(BaseService):
    EmailTemplate = DataEndpoint()
    MessageInstruction = DataEndpoint()
    MessageQueue = DataEndpoint()
    NotificationFilter = DataEndpoint()


@register
class PlayerService(BaseService):
    Player = Endpoint(name='p')


@register
class PlayerDataService(BaseService):
    AccountSettings = DataEndpoint()
    ColorScheme = DataEndpoint()
    Layout = DataEndpoint()
    Player = DataEndpoint()
    PlugIn = DataEndpoint()
    Skin = DataEndpoint()


@register
class ProductFeedsService(BaseService):
    Feed = Endpoint(name='')


@register
class ProductDataService(BaseService):
    AccountSettings = DataEndpoint()
    PricingTemplate = DataEndpoint()
    Product = DataEndpoint()
    ProductTag = DataEndpoint()
    Subscription = DataEndpoint()


@register
class PromotionDataService(BaseService):
    Promotion = DataEndpoint()
    PromotionAction = DataEndpoint()
    PromotionCode = DataEndpoint()
    PromotionCondition = DataEndpoint()
    PromotionUseCount = DataEndpoint()


@register
class PublishDataService(BaseService):
    Adapter = DataEndpoint()
    AdapterConfiguration = DataEndpoint()
    PublishProfile = DataEndpoint()


@register
class PublishService(BaseService):
    Publish = BusinessEndpoint()
    Sharing = BusinessEndpoint()


@register
class SelectorService(BaseService):
    Selector = Endpoint(name='')


@register
class SharingDataService(BaseService):
    OutletProfile = DataEndpoint()
    ProviderAdapter = DataEndpoint()


@register
class SocialDataService(BaseService):
    AccountSettings = DataEndpoint()
    Comment = DataEndpoint()
    Rating = DataEndpoint()
    TotalRating = DataEndpoint()


@register
class AdminStorefrontService(BaseService):
    Action = DataEndpoint()
    Contract = DataEndpoint()
    FulfillmentItem = DataEndpoint()
    Order = DataEndpoint()
    OrderFulfillment = DataEndpoint()
    OrderHistory = DataEndpoint()
    OrderItem = DataEndpoint()
    OrderSummary = DataEndpoint()
    PaymentInstrumentInfo = DataEndpoint()
    Shipment = DataEndpoint()
    Checkout = BusinessEndpoint(schema='1.5')
    Payment = BusinessEndpoint(schema='1.1')


@register
class StorefrontService(BaseService):
    Checkout = BusinessEndpoint(schema='1.4')
    Payment = BusinessEndpoint(schema='1.1')
    OrderHistory = DataEndpoint()
    PaymentInstrumentInfo = DataEndpoint()


@register
class TaskService(BaseService):
    Agent = DataEndpoint()
    Batch = DataEndpoint()
    Task = DataEndpoint()
    TaskTemplate = DataEndpoint()
    TaskType = DataEndpoint()


@register
class UserDataService(BaseService):
    Directory = DataEndpoint()
    Security = DataEndpoint()
    User = DataEndpoint()
    Authentication = BusinessEndpoint()
    Lookup = BusinessEndpoint()
    Self = BusinessEndpoint()


@register
class UserProfileDataService(BaseService):
    AccountSettings = DataEndpoint()
    TotalItem = DataEndpoint()
    UserList = DataEndpoint()
    UserListItem = DataEndpoint()
    UserProfile = DataEndpoint()


@register
class ValidationService(BaseService):
    Validation = BusinessEndpoint(schema='1.1')


@register
class ValidationDataService(BaseService):
    ConditionalRule = DataEndpoint()
    ValidationRule = DataEndpoint()
    Validator = DataEndpoint()


@register
class WatchFolderDataService(BaseService):
    registry_key = 'WatchFolder Data Service'
    WatchFolder = DataEndpoint()
    WatchFolderFile = DataEndpoint()


@register
class WorkflowDataService(BaseService):
    ProfileResult = DataEndpoint()
    ProfileStepResult = DataEndpoint()
    WorkflowQueue = DataEndpoint()


